# frozen_string_literal: true

module Hermes
  # Score class
  class Score
    include Hermes::Concerns::CommonFieldsConcern
    include Hermes::Concerns::ActivableConcern

    ORIENTATION_BOOLS = %w[
      single_parent
      priority_public
      mutation
    ].freeze

    # applicant weighting
    @@under_30_year_points = 60
    @@rejected_cal_points = 35
    @@declined_cal_points = -35

    # applicant entity weighting
    @@entity_in_peec_points = 210
    @@entity_goal_points = 210

    # priority_points
    @@priority_points_thresholds = [30, 70, 145]

    field :department, type: String
    field :sne_date, type: DateTime
    field :profile_score, type: Integer
    field :last_eligible_position, type: Integer
    field :epci, type: Integer
    field :profile_epci_score, type: Integer
    # score en comptant la position candidature
    field :position_points, type: Integer
    field :score_with_position, type: Integer
    field :has_applicated, type: Boolean, default: false
    field :eligible_position, type: Integer
    field :applicated_position, type: Integer
    field :effort_rate, type: Float
    field :priority_public, type: Boolean, default: false
    field :single_parent, type: Boolean, default: false
    field :mutation, type: Boolean, default: false
    field :reserved, type: Boolean, default: false
    field :entity_id, type: BSON::ObjectId

    common_indexes *%i[
      sne_date
      profile_score
      epci
      profile_epci_score
      position_points
      score_with_position
      has_applicated
      eligible_position
      applicated_position
      effort_rate
      priority_public
      single_parent
      mutation
      reserved
      entity_id
    ]

    belongs_to(
      :applicant,
      class_name: 'Hermes::Applicant',
      index: true
    )
    belongs_to(
      :housing_offer,
      class_name: 'Hermes::HousingOffer',
      index: true
    )

    validates_presence_of :applicant_id
    validates_presence_of :housing_offer_id

    scope :inactives, -> { where(is_active: false) }
    scope :applications, -> { where(has_applicated: true) }

    before_save :before_actions
    after_update :after_actions
    after_create :after_create_actions
    after_destroy :after_destroy_actions

    def calculate_eligible_position
      Hermes::Score.where(
        id: { :$ne => id },
        housing_offer_id: housing_offer_id,
        :$or => [
          { profile_epci_score: { :$gt => profile_epci_score } },
          {
            profile_epci_score: profile_epci_score,
            sne_date: { :$lt => sne_date }
          }
        ]
      ).count + 1
    end

    def update_eligible_position
      self.eligible_position = calculate_eligible_position
    end

    def calculate_applicated_position
      if Hermes::Score.where(
        has_applicated: true,
        id: { :$ne => id },
        housing_offer_id: housing_offer_id,
        :$or => [
          { profile_epci_score: { :$gt => profile_epci_score + 20 } },
          {
            profile_epci_score: profile_epci_score + 20,
            sne_date: { :$lt => sne_date }
          }
        ]
      ).count <= 20
        firsts = Hermes::Score.where(
          has_applicated: true,
          id: { :$ne => id },
          housing_offer_id: housing_offer_id
        ).order_by(score_with_position: -1).limit(20).to_a
        position_points = 20 - firsts.reject { |score| score.sne_date && (score.sne_date < sne_date) }.length
        Hermes::Score.where(
          has_applicated: true,
          id: { :$ne => id },
          housing_offer_id: housing_offer_id,
          :$or => [
            { profile_epci_score: { :$gt => profile_epci_score + position_points } },
            {
              profile_epci_score: profile_epci_score + position_points,
              sne_date: { :$lt => sne_date }
            }
          ]
        ).count + 1
      else
        21
      end
    end

    def update_applicated_position
      self.applicated_position = calculate_applicated_position
    end

    def active_request
      applicant&.first_active_request
    end

    def entity
      applicant&.entity
    end

    def monthly_revenues
      active_request&.monthly_revenues || 0.0
    end

    def rent_amount
      housing_offer&.rent_amount
    end

    def coapplicants
      return unless (request = active_request)

      request.coapplicants_raw || []
    end

    def any_applicant_is_under_30
      requesters_births = [applicant&.birth_date].concat(
          coapplicants.map { |coapplicant| coapplicant.birth_date }.to_a
        ).uniq.compact
      requesters_births.any? { |birth| birth + 30.years >= DateTime.current.utc }
    end

    def entity_points
      return 0 unless entity && (request = active_request)&.entity_validated

      points = 0
      points += @@entity_in_peec_points if entity.in_peec
      if request.priority_points?
        points += request.priority_points_indexes.inject(0) do |sum, index|
          sum + index
        end
      end

      points + @@entity_goal_points * (entity.achieved_goal || 0)
    end

    def calculate_effort_rate_points
      return 0 unless monthly_revenues.to_i.positive? && rent_amount.present?

      if (effort_rate = rent_amount.to_f / monthly_revenues) >= 0.26
        35
      else
        (((effort_rate.to_i * 100) / 26) * 35).to_i
      end
    end

    def already_rejected_in_cal
      active_request.rejected_in_cal_nb&.send('>', 0)
    end

    def already_declined_cal
      active_request.rejected_after_cal_nb&.send('>', 0)
    end

    def calculate_history_points
      score = 0
      score += @@rejected_cal_points if already_rejected_in_cal
      score += @@declined_cal_points if already_declined_cal
      score
    end

    def calculate_score
      return unless active_request && housing_offer

      score = 0
      score += @@under_30_year_points if any_applicant_is_under_30
      score += calculate_history_points

      score += entity_points

      score + calculate_effort_rate_points
    end

    def applicated?
      active_request&.housing_offer_ids
                    &.include?(housing_offer_id.to_s)
    end

    def update_has_applicated
      self.has_applicated = applicated? unless has_applicated_changed?
    end

    # set sne_date to the oldest of nuds creation_date matching the offer's department
    def update_sne_date
      self.sne_date = active_request&.nuds
                                    &.where(department: housing_offer.department)
                                    &.order_by(creation_date: 1)
                                    &.first
                                    &.creation_date
    end

    def update_profile_score
      self.profile_score = calculate_score
    end

    def process
      save
    end

    def update_department
      self.department = housing_offer.department
    end

    def update_effort_rate
      self.effort_rate = calculate_effort_rate_points
    end

    def update_profile_epci_score
      self.profile_epci_score = (profile_score || 0) + (epci || 0)
    end

    def update_score_with_position
      return unless (position = applicated_position - 1) < 20

      pos = (inter = 20 - position).positive? ? inter : 0
      self.position_points = pos
      self.score_with_position = profile_epci_score + pos
    end

    def inherit_data
      if active_request
        ORIENTATION_BOOLS.each do |orientation_name|
          send(
            "#{orientation_name}=",
            active_request.send(orientation_name)
          )
        end
        self.reserved = active_request.entity_validated
      end
      self.entity_id = entity&.id if reserved
    end

    def before_actions
      inherit_data
      update_has_applicated
      update_department
      update_effort_rate
      update_profile_score
      update_profile_epci_score
      update_sne_date
      if profile_epci_score_changed?
        update_eligible_position
        update_applicated_position
      end
      update_has_applicated if has_applicated_changed?
      update_score_with_position if has_applicated_changed? && has_applicated
    end

    def handle_others_eligible
      UpdateEligibleJob.perform_later(id.to_s)
    end

    def handle_applications
      Hermes::Score.where(
        id: { :$ne => id },
        has_applicated: true,
        sne_date: { :lt => sne_date }
      ).each do |lower_score|
        new_points = lower_score.position_points - 1
        lower_score.position_points = new_points.positive? ? new_points : nil
        lower_score.save
      end
    end

    def orientation_status
      orientation = ORIENTATION_BOOLS.each_with_object({}) do |orientation_name, h|
        h[orientation_name.to_sym] = if send(orientation_name)
                                       1
                                     else
                                       -1
                                     end
      end
      orientation[:entities] = { entity_id.to_s => 1 } if reserved
      orientation
    end

    def inverted_orientation_status
      orientation_status.inverted_values
    end

    def unnegated_inverted_orientation_status
      orientation_status.unnegate.inverted_values
    end

    def orientation_changes
      changes = ORIENTATION_BOOLS.each_with_object({}) do |orientation_name, h|
        if send("#{orientation_name}_changed?")
          h[orientation_name.to_sym] = if send(orientation_name)
                                         -1
                                       else
                                         1
                                       end
        end
      end
      if reserved_changed?
        changes[:entities] = { entity_id.to_s => reserved ? 1 : -1 }
      elsif reserved && entity_id_changed?
        changes[:entities] = {
          entity_id_was.to_s => -1,
          entity_id.to_s => 1
        }
      end
      changes
    end

    def orientation_changed?
      orientation_changes.present?
    end

    def handle_orientation_movements(for_deactivation = false)
      return unless housing_offer_id

      changes = if for_deactivation
                  unnegated_inverted_orientation_status
                elsif created_at == updated_at || new_record?
                  orientation_status.unnegate
                else
                  orientation_changes
                end
      HandleOfferEligibleNbJob.perform_later(housing_offer_id.to_s, changes)
    end

    def after_actions
      if eligible_position_changed?
        self.set(last_eligible_position:self.eligible_position_was)
        handle_others_eligible
      end
      if is_active_changed? && !is_active
        handle_orientation_movements(true)
      elsif orientation_changed?
        handle_orientation_movements
      end
      handle_applications if has_applicated_changed? && has_applicated
    end

    def after_destroy_actions
      handle_orientation_movements(true)
    end

    def after_create_actions
      after_actions
      Hermes::Log.create(
        code: 'SC_C',
        action: "new score_created with applicant #{applicant_id} and housing_offer #{housing_offer_id}",
        score_id: id
      )
    end

    class << self
      # A utiliser dans HousingOfferSerializer
      def retrieve(applicant_id, housing_offer_id)
        s = Score.find_by(applicant_id: applicant_id, housing_offer_id: housing_offer_id )
        unless s
          s = Score.create(applicant_id: applicant_id, housing_offer_id: housing_offer_id )
          s.process
        end
        s
      end
    end
  end
end
